import json
import logging
import datetime
import dateutil
import pandas as pd

from ..calculator import Calculator
from ..utils import get_extension, to_json_string, StatusTypes

logger = logging.getLogger(__name__)

class CycleTimeCalculator(Calculator):
    """Basic cycle time data, fetched from JIRA.

    Builds a numerically indexed data frame with the following 'fixed'
    columns: `key`, 'url', 'issue_type', `summary`, `status`, and
    `resolution` from JIRA, as well as the value of any fields set in
    the `fields` dict in `settings`. If `known_values` is set (a dict of
    lists, with field names as keys and a list of known values for each
    field as values) and a field in `fields` contains a list of values,
    only the first value in the list of known values will be used.

    If 'query_attribute' is set in `settings`, a column with this name
    will be added, and populated with the `value` key, if any, from each
    criteria block under `queries` in settings.

    In addition, `cycle_time` will be set to the time delta between the
    first `accepted`-type column and the first `complete` column, or None.

    The remaining columns are the names of the items in the configured
    cycle, in order.

    Each cell contains the last date/time stamp when the relevant status
    was set.

    If an item moves backwards through the cycle, subsequent date/time
    stamps in the cycle are erased.
    """

    def __init__(self, query_manager, settings, results):
        super().__init__(query_manager, settings, results)

        self.cycle_lookup = {}
        for idx, cycle_step in enumerate(self.settings['cycle']):
            for status in cycle_step['statuses']:
                self.cycle_lookup[status.lower()] = dict(
                    index=idx,
                    name=cycle_step['name'],
                    type=cycle_step['type'],
                )

    def run(self, now=None):

        # Allows unit testing to use a fixed date
        if now is None:
            now = datetime.datetime.utcnow()
        
        cycle_names = [s['name'] for s in self.settings['cycle']]
        accepted_steps = set(s['name'] for s in self.settings['cycle'] if s['type'] == StatusTypes.accepted)
        completed_steps = set(s['name'] for s in self.settings['cycle'] if s['type'] == StatusTypes.complete)

        backlog_column = self.settings['backlog_column']
        done_column = self.settings['done_column']

        series = {
            'key': {'data': [], 'dtype': 'str'},
            'url': {'data': [], 'dtype': 'str'},
            'issue_type': {'data': [], 'dtype': 'str'},
            'summary': {'data': [], 'dtype': 'str'},
            'status': {'data': [], 'dtype': 'str'},
            'resolution': {'data': [], 'dtype': 'str'},
            'cycle_time': {'data': [], 'dtype': 'timedelta64[ns]'},
            'completed_timestamp': {'data': [], 'dtype': 'datetime64[ns]'},
            'blocked_days': {'data': [], 'dtype': 'int'},
            'impediments': {'data': [], 'dtype': 'object'},  # list of {'start', 'end', 'status'}
        }

        for cycle_name in cycle_names:
            series[cycle_name] = {'data': [], 'dtype': 'datetime64[ns]'}

        for name in self.settings['attributes'].keys():
            series[name] = {'data': [], 'dtype': 'object'}

        if self.settings['query_attribute']:
            series[self.settings['query_attribute']] = {'data': [], 'dtype': 'str'}

        for criteria in self.settings['queries']:
            for issue in self.query_manager.find_issues(criteria['jql']):

                item = {
                    'key': issue.key,
                    'url': "%s/browse/%s" % (self.query_manager.jira._options['server'], issue.key,),
                    'issue_type': issue.fields.issuetype.name,
                    'summary': issue.fields.summary,
                    'status': issue.fields.status.name,
                    'resolution': issue.fields.resolution.name if issue.fields.resolution else None,
                    'cycle_time': None,
                    'completed_timestamp': None,
                    'blocked_days': 0,
                    'impediments': []
                }

                for name in self.settings['attributes'].keys():
                    item[name] = self.query_manager.resolve_attribute_value(issue, name)

                if self.settings['query_attribute']:
                    item[self.settings['query_attribute']] = criteria.get('value', None)

                for cycle_name in cycle_names:
                    item[cycle_name] = None

                last_status = None
                impediment_start_status = None
                impediment_start = None

                # Record date of status and impediments flag changes
                for snapshot in self.query_manager.iter_changes(issue, ['status', 'Flagged']):
                    if snapshot.change == 'status':
                        snapshot_cycle_step = self.cycle_lookup.get(snapshot.to_string.lower(), None)
                        if snapshot_cycle_step is None:
                            logger.warn("Issue %s transitioned to unknown JIRA status %s", issue.key, snapshot.to_string)
                            continue
                        
                        last_status = snapshot_cycle_step_name = snapshot_cycle_step['name']

                        # Keep the first time we entered a step
                        if item[snapshot_cycle_step_name] is None:
                            item[snapshot_cycle_step_name] = snapshot.date

                        # Wipe any subsequent dates, in case this was a move backwards
                        found_cycle_name = False
                        for cycle_name in cycle_names:
                            if not found_cycle_name and cycle_name == snapshot_cycle_step_name:
                                found_cycle_name = True
                                continue
                            elif found_cycle_name and item[cycle_name] is not None:
                                logger.info("Issue %s moved backwards to %s, wiping data for subsequent step %s", issue.key, snapshot_cycle_step_name, cycle_name)
                                item[cycle_name] = None
                    elif snapshot.change == 'Flagged':
                        if snapshot.from_string == snapshot.to_string is None:
                            # Initial state from None -> None
                            continue
                        elif snapshot.to_string is not None and snapshot.to_string != "":
                            impediment_start = snapshot.date.date()
                            impediment_start_status = last_status
                        elif snapshot.to_string is None or snapshot.to_string == "":
                            if impediment_start is None:
                                logger.warning("Issue %s had impediment flag cleared before being set. This should not happen.", issue.key)
                                continue
                            
                            if impediment_start_status not in (backlog_column, done_column):
                                item['blocked_days'] += (snapshot.date.date() - impediment_start).days
                            item['impediments'].append({'start': impediment_start, 'end': snapshot.date.date(), 'status': impediment_start_status})

                            # Reset for next time
                            impediment_start = None
                            impediment_start_status = None
                
                # If an impediment flag was set but never cleared: treat as resolved on the ticket
                # resolution date if the ticket was resolved, else as still open until today.
                if impediment_start is not None:
                    if issue.fields.resolutiondate:
                        resolution_date = dateutil.parser.parse(issue.fields.resolutiondate).date()
                        if impediment_start_status not in (backlog_column, done_column):
                            item['blocked_days'] += (resolution_date - impediment_start).days
                        item['impediments'].append({'start': impediment_start, 'end': resolution_date, 'status': impediment_start_status})
                    else:
                        if impediment_start_status not in (backlog_column, done_column):
                            item['blocked_days'] += (now.date() - impediment_start).days
                        item['impediments'].append({'start': impediment_start, 'end': None, 'status': impediment_start_status})
                    impediment_start = None
                    impediment_start_status = None
                
                # Wipe timestamps if items have moved backwards; calculate cycle time

                previous_timestamp = None
                accepted_timestamp = None
                completed_timestamp = None

                for cycle_name in cycle_names:
                    if item[cycle_name] is not None:
                        previous_timestamp = item[cycle_name]

                        if accepted_timestamp is None and previous_timestamp is not None and cycle_name in accepted_steps:
                            accepted_timestamp = previous_timestamp
                        if completed_timestamp is None and previous_timestamp is not None and cycle_name in completed_steps:
                            completed_timestamp = previous_timestamp

                if accepted_timestamp is not None and completed_timestamp is not None:
                    item['cycle_time'] = completed_timestamp - accepted_timestamp
                    item['completed_timestamp'] = completed_timestamp

                for k, v in item.items():
                    series[k]['data'].append(v)

        data = {}
        for k, v in series.items():
            data[k] = pd.Series(v['data'], dtype=v['dtype'])

        return pd.DataFrame(data,
            columns=['key', 'url', 'issue_type', 'summary', 'status', 'resolution'] +
                    sorted(self.settings['attributes'].keys()) +
                    ([self.settings['query_attribute']] if self.settings['query_attribute'] else []) +
                    ['cycle_time', 'completed_timestamp', 'blocked_days', 'impediments'] +
                    cycle_names
        )

    def write(self):
        output_file = self.settings['cycle_time_data']
        
        if not output_file:
            logger.debug("No output file specified for cycle time data")
            return

        output_extension = get_extension(output_file)

        cycle_data = self.get_result()
        cycle_names = [s['name'] for s in self.settings['cycle']]
        attribute_names = sorted(self.settings['attributes'].keys())
        query_attribute_names = [self.settings['query_attribute']] if self.settings['query_attribute'] else []

        header = ['ID', 'Link', 'Name'] + cycle_names + ['Type', 'Status', 'Resolution'] + attribute_names + query_attribute_names + ['Blocked Days']
        columns = ['key', 'url', 'summary'] + cycle_names + ['issue_type', 'status', 'resolution'] + attribute_names + query_attribute_names + ['blocked_days']

        logger.info("Writing cycle time data to %s", output_file)

        if output_extension == '.json':
            values = [header] + [list(map(to_json_string, row)) for row in cycle_data[columns].values.tolist()]
            with open(output_file, 'w') as out:
                out.write(json.dumps(values))
        elif output_extension == '.xlsx':
            cycle_data.to_excel(output_file, 'Cycle data', columns=columns, header=header, index=False)
        else:
            cycle_data.to_csv(output_file, columns=columns, header=header, date_format='%Y-%m-%d', index=False)



import pytest
from datetime import date
from pandas import DataFrame, NaT, Timestamp

from .cycletime import CycleTimeCalculator
from .impediments import ImpedimentsCalculator

from ..utils import extend_dict

from ..conftest import _issues

def _ts(datestring, timestring="00:00:00", freq=None):
    return Timestamp('%s %s' % (datestring, timestring,), freq=freq)

@pytest.fixture
def query_manager(minimal_query_manager):
    return minimal_query_manager

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {
        'impediments_data': 'impediments.csv',
        'impediments_chart': 'impediments.png',
        'impediments_days_chart': 'impediments-days.png',
    })

@pytest.fixture
def columns(minimal_cycle_time_columns):
    return minimal_cycle_time_columns

@pytest.fixture
def cycle_time_results(minimal_cycle_time_columns):
    """A results dict mimicing a minimal result from the CycleTimeCalculator.
    """
    return {
        CycleTimeCalculator: DataFrame(_issues([
            dict(Backlog=_ts('2018-01-01'), Committed=NaT, Build=NaT, Test=NaT, Done=NaT, blocked_days=0, impediments=[]),
            dict(Backlog=_ts('2018-01-02'), Committed=_ts('2018-01-03'), Build=NaT, Test=NaT, Done=NaT, blocked_days=4, impediments=[
                {'start': date(2018, 1, 5),  'end': date(2018, 1, 7),  'status': 'Backlog'},      # ignored because it was blocked in backlog
                {'start': date(2018, 1, 10), 'end': date(2018, 1, 12), 'status': 'Committed'},  # included
            ]),
            dict(Backlog=_ts('2018-01-03'), Committed=_ts('2018-01-03'), Build=_ts('2018-01-04'), Test=_ts('2018-01-05'), Done=_ts('2018-01-06'), blocked_days=4, impediments=[
                {'start': date(2018, 1, 4), 'end': date(2018, 1, 5),  'status': 'Build'},  # included
                {'start': date(2018, 1, 7), 'end': date(2018, 1, 10), 'status': 'Done'},  # ignored because it was blocked in done
            ]),
            dict(Backlog=_ts('2018-01-04'), Committed=_ts('2018-01-04'), Build=NaT, Test=NaT, Done=NaT, blocked_days=100, impediments=[
                {'start': date(2018, 1, 5), 'end': None, 'status': 'Committed'},  # open ended, still included
            ]),
        ]), columns=minimal_cycle_time_columns)
    }
    
def test_only_runs_if_charts_set(query_manager, settings, cycle_time_results):
    test_settings = extend_dict(settings, {
        'impediments_data': None,
        'impediments_chart': None,
        'impediments_days_chart': None,
    })

    calculator = ImpedimentsCalculator(query_manager, test_settings, cycle_time_results)
    data = calculator.run()
    assert data is None

    test_settings = extend_dict(settings, {
        'impediments_data': 'impediments.csv',
        'impediments_chart': None,
        'impediments_days_chart': None,
    })

    calculator = ImpedimentsCalculator(query_manager, test_settings, cycle_time_results)
    data = calculator.run()
    assert data is not None

    test_settings = extend_dict(settings, {
        'impediments_data': None,
        'impediments_chart': 'impediments.png',
        'impediments_days_chart': None,
    })

    calculator = ImpedimentsCalculator(query_manager, test_settings, cycle_time_results)
    data = calculator.run()
    assert data is not None

    test_settings = extend_dict(settings, {
        'impediments_data': None,
        'impediments_chart': None,
        'impediments_days_chart': 'days.png',
    })

    calculator = ImpedimentsCalculator(query_manager, test_settings, cycle_time_results)
    data = calculator.run()
    assert data is not None

def test_empty(query_manager, settings, columns):
    results = {
        CycleTimeCalculator: DataFrame([], columns=columns)
    }

    calculator = ImpedimentsCalculator(query_manager, settings, results)

    data = calculator.run()
    assert len(data.index) == 0

def test_columns(query_manager, settings, cycle_time_results):
    calculator = ImpedimentsCalculator(query_manager, settings, cycle_time_results)

    data = calculator.run()

    assert list(data.columns) == ['key', 'status', 'start', 'end']

def test_calculate_impediments(query_manager, settings, cycle_time_results):
    calculator = ImpedimentsCalculator(query_manager, settings, cycle_time_results)

    data = calculator.run()

    assert data.to_dict('records') == [
        {'key': 'A-2', 'status': 'Committed', 'start': _ts('2018-01-10'), 'end': _ts('2018-01-12')},
        {'key': 'A-3', 'status': 'Build',     'start': _ts('2018-01-04'), 'end': _ts('2018-01-05')},
        {'key': 'A-4', 'status': 'Committed', 'start': _ts('2018-01-05'), 'end': NaT},
    ]

def test_different_backlog_column(query_manager, settings, cycle_time_results):
    settings = extend_dict(settings, {
        'backlog_column': 'Committed',
    })
    calculator = ImpedimentsCalculator(query_manager, settings, cycle_time_results)

    data = calculator.run()

    assert data.to_dict('records') == [
        {'key': 'A-2', 'status': 'Backlog', 'start': _ts('2018-01-05'), 'end': _ts('2018-01-07')},
        {'key': 'A-3', 'status': 'Build',   'start': _ts('2018-01-04'), 'end': _ts('2018-01-05')},
    ]

def test_different_done_column(query_manager, settings, cycle_time_results):
    settings = extend_dict(settings, {
        'done_column': 'Build',
    })
    calculator = ImpedimentsCalculator(query_manager, settings, cycle_time_results)

    data = calculator.run()

    assert data.to_dict('records') == [
        {'key': 'A-2', 'status': 'Committed', 'start': _ts('2018-01-10'), 'end': _ts('2018-01-12')},
        {'key': 'A-3', 'status': 'Done',      'start': _ts('2018-01-07'), 'end': _ts('2018-01-10')},
        {'key': 'A-4', 'status': 'Committed', 'start': _ts('2018-01-05'), 'end': NaT},
    ]



import pytest
from pandas import DataFrame

from .cycletime import CycleTimeCalculator
from .histogram import HistogramCalculator

from ..utils import extend_dict

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {})

@pytest.fixture
def query_manager(minimal_query_manager):
    return minimal_query_manager

@pytest.fixture
def results(large_cycle_time_results):
    return extend_dict(large_cycle_time_results, {})

def test_empty(query_manager, settings, minimal_cycle_time_columns):
    results = {
        CycleTimeCalculator: DataFrame([], columns=minimal_cycle_time_columns, index=[])
    }

    calculator = HistogramCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.index) == [
        '0.0 to 0.1',
        '0.1 to 0.2',
        '0.2 to 0.3',
        '0.3 to 0.4',
        '0.4 to 0.5',
        '0.5 to 0.6',
        '0.6 to 0.7',
        '0.7 to 0.8',
        '0.8 to 0.9',
        '0.9 to 1.0'
    ]

    assert list(data) == [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

def test_calculate_histogram(query_manager, settings, results):
    calculator = HistogramCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.index) == [
        '4.0 to 4.1',
        '4.1 to 4.2',
        '4.2 to 4.3',
        '4.3 to 4.4',
        '4.4 to 4.5',
        '4.5 to 4.6',
        '4.6 to 4.7',
        '4.7 to 4.8',
        '4.8 to 4.9',
        '4.9 to 5.0'
    ]
    assert list(data) == [1, 0, 0, 0, 0, 0, 0, 0, 0, 5]



import json

from .cli import (
    override_options
)

def test_override_options():

    class FauxArgs:
        def __init__(self, opts):
            self.__dict__.update(opts)
            for k, v in opts.items():
                setattr(self, k, v)

    options = {'one': 1, 'two': 2}
    override_options(options, FauxArgs({}))
    assert json.dumps(options) == json.dumps({'one': 1, 'two': 2})
    
    options = {'one': 1, 'two': 2}
    override_options(options, FauxArgs({'one': 11}))
    assert json.dumps(options) == json.dumps({'one': 11, 'two': 2})

    options = {'one': 1, 'two': 2}
    override_options(options, FauxArgs({'three': 3}))
    assert json.dumps(options) == json.dumps({'one': 1, 'two': 2})



import logging
import pandas as pd
import matplotlib.pyplot as plt

from ..calculator import Calculator
from ..utils import get_extension, breakdown_by_month, breakdown_by_month_sum_days, set_chart_style

from .cycletime import CycleTimeCalculator

logger = logging.getLogger(__name__)

class ImpedimentsCalculator(Calculator):
    """Calculate impediments, charted by month and workflow status, either as
    a count of tickets that were blocked in that month, or as a sum of the total
    number of days of blockage for all tickets in that month.

    Writes to `impediments_chart` and `impediments_days_chart`, respectively,
    with corresponding titles. The number of months to output can be restricted
    with `impediments_window`. Raw data can be written to `impediments_data`.
    """

    def run(self):

        # This calculation is expensive. Only run it if we are going to write something
        if not (
            self.settings['impediments_data'] or
            self.settings['impediments_chart'] or
            self.settings['impediments_days_chart']
        ):
            logger.debug("Not calculating impediments data as no output files specified")
            return None
        
        backlog_column = self.settings['backlog_column']
        done_column = self.settings['done_column']

        cycle_data = self.get_result(CycleTimeCalculator)
        cycle_data = cycle_data[cycle_data.blocked_days > 0][['key', 'impediments']]
        
        data = []

        for row in cycle_data.itertuples():
            for idx, event in enumerate(row.impediments):
                # Ignore things that were impeded whilst in the backlog and/or done column
                # (these are mostly nonsensical, and don't really indicate blocked/wasted time)

                if event['status'] in (backlog_column, done_column):
                    continue
                data.append({
                    'key': row.key,
                    'status': event['status'],
                    'start': pd.Timestamp(event['start']),
                    'end': pd.Timestamp(event['end']) if event['end'] else pd.NaT,
                })
        
        return pd.DataFrame(data, columns=['key', 'status', 'start', 'end'])

    def write(self):
        data = self.get_result()
        if data is None:
            return

        if self.settings['impediments_data']:
            self.write_data(data, self.settings['impediments_data'])

        if self.settings['impediments_chart']:
            self.write_impediments_chart(data, self.settings['impediments_chart'])
        
        if self.settings['impediments_days_chart']:
            self.write_impediments_days_chart(data, self.settings['impediments_days_chart'])
    
    def write_data(self, data, output_file):
        output_extension = get_extension(output_file)

        logger.info("Writing impediments data to %s", output_file)
        if output_extension == '.json':
            data.to_json(output_file, date_format='iso')
        elif output_extension == '.xlsx':
            data.to_excel(output_file, 'Impediments', header=True)
        else:
            data.to_csv(output_file, header=True, date_format='%Y-%m-%d', index=False)

    def write_impediments_chart(self, chart_data, output_file):
        if len(chart_data.index) == 0:
            logger.warning("Cannot draw impediments chart with zero items")
            return
        
        window = self.settings['impediments_window']
        cycle_names = [s['name'] for s in self.settings['cycle']]

        breakdown = breakdown_by_month(chart_data, 'start', 'end', 'key', 'status', cycle_names)
        
        if window:
            breakdown = breakdown[-window:]

            if len(breakdown.index) == 0:
                logger.warning("Cannot draw impediments chart with zero items")
                return

        fig, ax = plt.subplots()
        
        breakdown.plot.bar(ax=ax, stacked=True)
        
        if self.settings['impediments_chart_title']:
            ax.set_title(self.settings['impediments_chart_title'])

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        ax.set_xlabel("Month", labelpad=20)
        ax.set_ylabel("Number of impediments", labelpad=10)

        labels = [d.strftime("%b %y") for d in breakdown.index]
        ax.set_xticklabels(labels, rotation=90, size='small')

        set_chart_style()

        # Write file
        logger.info("Writing impediments chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)
    
    def write_impediments_days_chart(self, chart_data, output_file):
        if len(chart_data.index) == 0:
            logger.warning("Cannot draw impediments days chart with zero items")
            return

        window = self.settings['impediments_window']
        cycle_names = [s['name'] for s in self.settings['cycle']]

        breakdown = breakdown_by_month_sum_days(chart_data, 'start', 'end', 'status', cycle_names)
        
        if window:
            breakdown = breakdown[-window:]

            if len(breakdown.index) == 0:
                logger.warning("Cannot draw impediments chart with zero items")
                return

        fig, ax = plt.subplots()
        
        breakdown.plot.bar(ax=ax, stacked=True)
        
        if self.settings['impediments_days_chart_title']:
            ax.set_title(self.settings['impediments_days_chart_title'])

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        ax.set_xlabel("Month", labelpad=20)
        ax.set_ylabel("Total impeded days", labelpad=10)

        labels = [d.strftime("%b %y") for d in breakdown.index]
        ax.set_xticklabels(labels, rotation=90, size='small')

        set_chart_style()

        # Write file
        logger.info("Writing impediments days chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



import logging
import pandas as pd
import matplotlib.pyplot as plt

from ..calculator import Calculator
from ..utils import set_chart_style

from .cfd import CFDCalculator

logger = logging.getLogger(__name__)

class BurnupCalculator(Calculator):
    """Draw a simple burn-up chart.
    """

    def run(self):
        cfd_data = self.get_result(CFDCalculator)
        
        backlog_column = self.settings['backlog_column']
        done_column = self.settings['done_column']

        if backlog_column not in cfd_data.columns:
            logger.error("Backlog column %s does not exist", backlog_column)
            return None
        if done_column not in cfd_data.columns:
            logger.error("Done column %s does not exist", done_column)
            return None

        return cfd_data[[backlog_column, done_column]]
    
    def write(self):
        output_file = self.settings['burnup_chart']
        if not output_file:
            logger.debug("No output file specified for burnup chart")
            return

        chart_data = self.get_result()

        if len(chart_data.index) == 0:
            logger.warning("Unable to draw burnup chart with no data items")
            return
        
        window = self.settings['burnup_window']
        if window:
            start = chart_data.index.max() - pd.Timedelta(window, 'D')
            chart_data = chart_data[start:]

            # Re-check after slicing for window
            if len(chart_data.index) == 0:
                logger.warning("Unable to draw burnup chart with no data items")
                return

        fig, ax = plt.subplots()
        
        if self.settings['burnup_chart_title']:
            ax.set_title(self.settings['burnup_chart_title'])

        fig.autofmt_xdate()

        ax.set_xlabel("Date")
        ax.set_ylabel("Number of items")

        chart_data.plot.line(ax=ax, legend=True)

        bottom = chart_data[chart_data.columns[-1]].min()
        top = chart_data[chart_data.columns[0]].max()
        ax.set_ylim(bottom=bottom, top=top)
        
        # Place legend underneath graph
        box = ax.get_position()
        handles, labels = ax.get_legend_handles_labels()
        ax.set_position([box.x0, box.y0 + box.height * 0.1, box.width, box.height * 0.9])

        ax.legend(handles[:2], labels[:2], loc='upper center', bbox_to_anchor=(0.5, -0.2), ncol=2)

        set_chart_style()

        # Write file
        logger.info("Writing burnup chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



import logging
import matplotlib.pyplot as plt
import pandas as pd

from ..calculator import Calculator
from ..utils import set_chart_style

from .cfd import CFDCalculator

logger = logging.getLogger(__name__)

class WIPChartCalculator(Calculator):
    """Draw a weekly WIP chart
    """

    def run(self):
        cfd_data = self.get_result(CFDCalculator)
        cycle_names = [s['name'] for s in self.settings['cycle']]

        start_column = self.settings['committed_column']
        done_column = self.settings['done_column']

        if start_column not in cycle_names:
            logger.error("Committed column %s does not exist", start_column)
            return None
        if done_column not in cycle_names:
            logger.error("Done column %s does not exist", done_column)
            return None
        
        return pd.DataFrame({'wip': cfd_data[start_column] - cfd_data[done_column]}, index=cfd_data.index)
    
    def write(self):
        output_file = self.settings['wip_chart']
        if not output_file:
            logger.debug("No output file specified for WIP chart")
            return

        chart_data = self.get_result()

        if len(chart_data.index) == 0:
            logger.warning("Cannot draw WIP chart with no completed items")
            return

        fig, ax = plt.subplots()
        
        if self.settings['wip_chart_title']:
            ax.set_title(self.settings['wip_chart_title'])

        frequency = self.settings['wip_frequency']
        logger.debug("Calculating WIP chart with frequency %s", frequency)

        wip_data = chart_data[['wip']]

        window = self.settings['wip_window']
        if window:
            start = wip_data.index.max().normalize() - (window * pd.tseries.frequencies.to_offset(frequency))
            wip_data = wip_data[start:]

            if len(wip_data.index) < 2:
                logger.warning("Need at least 2 completed items to draw scatterplot")
                return

        groups = wip_data.groupby(pd.Grouper(freq=frequency, label='left', closed='left'))
        labels = [x[0].strftime("%d/%m/%Y") for x in groups]

        groups.boxplot(subplots=False, ax=ax, showmeans=True, return_type='axes')
        ax.set_xticklabels(labels, rotation=70, size='small')

        ax.set_xlabel("Period starting")
        ax.set_ylabel("WIP")

        set_chart_style()

        # Write file
        logger.info("Writing WIP chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



from .calculator import (
    Calculator,
    run_calculators
)

def test_run_calculator():
    
    written = []

    class Enabled(Calculator):

        def run(self):
            return "Enabled"
        
        def write(self):
            written.append("Enabled")
    
    class Disabled(Calculator):

        def run(self):
            return "Disabled"
        
        def write(self):
            pass

    class GetPreviousResult(Calculator):

        def run(self):
            return self.get_result(Enabled) + " " + self.settings['foo']
        
        def write(self):
            written.append(self.get_result())
    
    calculators = [Enabled, Disabled, GetPreviousResult]
    query_manager = object()
    settings = {'foo': 'bar'}

    results = run_calculators(calculators, query_manager, settings)

    assert results == {
        Enabled: "Enabled",
        Disabled: "Disabled",
        GetPreviousResult: "Enabled bar"
    }

    assert written == ["Enabled", "Enabled bar"]



import itertools
import logging
import dateutil.parser
import dateutil.tz

from .config import ConfigError

logger = logging.getLogger(__name__)

class IssueSnapshot(object):
    """A snapshot of the key fields of an issue at a point in its change history
    """

    def __init__(self, change, key, date, from_string, to_string):
        self.change = change
        self.key = key
        self.date = date.astimezone(dateutil.tz.tzutc())
        self.from_string = from_string
        self.to_string = to_string

    def __eq__(self, other):
        return all((
            self.change == other.change,
            self.key == other.key,
            self.date.isoformat() == other.date.isoformat(),
            self.from_string == other.from_string,
            self.to_string == other.to_string
        ))

    def __repr__(self):
        return "<IssueSnapshot change=%s key=%s date=%s from=%s to=%s>" % (
            self.change, self.key, self.date.isoformat(), self.from_string, self.to_string,
        )

class QueryManager(object):
    """Manage and execute queries
    """

    settings = dict(
        attributes={},
        known_values={},
        max_results=False,
    )

    def __init__(self, jira, settings):
        self.jira = jira
        self.settings = self.settings.copy()
        self.settings.update(settings)

        self.attributes_to_fields = {}
        self.fields_to_attributes = {}

        # Look up fields in JIRA and resolve attributes to fields
        logger.debug("Resolving JIRA fields")
        self.jira_fields = self.jira.fields()
        field_id = None

        for name, field in self.settings['attributes'].items():
            field_id = self.field_name_to_id(field)
            self.attributes_to_fields[name] = field_id
            self.fields_to_attributes[field_id] = name

    def field_name_to_id(self, name):
        try:
            return next((f['id'] for f in self.jira_fields if f['name'].lower() == name.lower()))
        except StopIteration:
            raise ConfigError("JIRA field with name `%s` does not exist (did you try to use the field id instead?)" % name) from None

    def resolve_attribute_value(self, issue, attribute_name):
        """Given an attribute name (i.e. one named in the config file and
        mapped to a field in JIRA), return its value from the given issue.
        Respects the `Known Values` settings and tries to resolve complex
        data types.
        """
        field_id = self.attributes_to_fields[attribute_name]
        return self.resolve_field_value(issue, field_id)

    def resolve_field_value(self, issue, field_id):
        """Given a JIRA internal field id, return its value from the given
        issue. Respects the `Known Values` settings and tries to resolve
        complex data types.
        """

        field_value = getattr(issue.fields, field_id)

        if field_value is None:
            return None

        value = getattr(field_value, 'value', field_value)

        if isinstance(value, (list, tuple)):
            if len(value) == 0:
                value = None
            else:
                values = [getattr(v, 'name', v) for v in value]

                # is this a `Known Values` attribute?
                attribute_name = self.fields_to_attributes.get(field_id, None)
                if attribute_name not in self.settings['known_values']:
                    value = values[0]
                else:
                    try:
                        value = next(filter(lambda v: v in values, self.settings['known_values'][attribute_name]))
                    except StopIteration:
                        value = None

        if not isinstance(value, (int, float, bool, str, bytes)):
            try:
                value = str(value)
            except TypeError:
                pass

        return value

    def iter_changes(self, issue, fields):
        """Yield an IssueSnapshot for each time the issue changed, including an
        initial value. `fields` is a list of fields to monitor, e.g.
        `['status']`.
        """

        for field in fields:
            initial_value = self.resolve_field_value(issue, self.field_name_to_id(field))
            try:
                initial_value = next(filter(
                    lambda h: h.field == field,
                    itertools.chain.from_iterable([c.items for c in issue.changelog.histories])
                )).fromString
            except StopIteration:
                pass
            
            yield IssueSnapshot(
                change=field,
                key=issue.key,
                date=dateutil.parser.parse(issue.fields.created),
                from_string=None,
                to_string=initial_value,
            )

        for change in issue.changelog.histories:
            change_date = dateutil.parser.parse(change.created)

            for item in change.items:
                if item.field in fields:
                    yield IssueSnapshot(
                        change=item.field,
                        key=issue.key,
                        date=change_date,
                        from_string=item.fromString,
                        to_string=item.toString
                    )

    # Basic queries

    def find_issues(self, jql, expand='changelog'):
        """Return a list of issues with changelog metadata for the given
        JQL.
        """

        max_results = self.settings['max_results']

        logger.info("Fetching issues with query `%s`", jql)
        if max_results:
            logger.info("Limiting to %d results", max_results)

        issues = self.jira.search_issues(jql, expand=expand, maxResults=max_results)
        logger.info("Fetched %d issues", len(issues))

        return issues



import datetime
import os.path

import numpy as np
import pandas as pd
import seaborn as sns

class StatusTypes:
    backlog = 'backlog'
    accepted = 'accepted'
    complete = 'complete'

def extend_dict(d, e):
    r = d.copy()
    r.update(e)
    return r

def to_json_string(value):
    if isinstance(value, pd.Timestamp):
        return value.strftime("%Y-%m-%d")
    if value in (None, np.NaN, pd.NaT):
        return ""

    try:
        return str(value)
    except TypeError:
        return value

def get_extension(filename):
    return os.path.splitext(filename)[1].lower()

def to_days_since_epoch(d):
    return (d - datetime.date(1970, 1, 1)).days

def set_chart_context(context):
    sns.set_context(context)

def set_chart_style(style="whitegrid", despine=True):
    sns.set_style(style)
    if despine:
        sns.despine()

def breakdown_by_month(df, start_column, end_column, key_column, value_column, output_columns=None, aggfunc='count'):
    """If `df` is a DataFrame of items that are valid/active between the
    timestamps stored in `start_column` and `end_column`, and where each item
    is uniquely identified by `key_column` and has a categorical value in
    `value_column`, return a new DataFrame counting the number of items in
    each month broken down by each unique value in `value_column`. To restrict
    (and order) the value columns, pass a list of valid values as `output_columns`.
    """
    
    def build_df(t):
        start_date = getattr(t, start_column)
        end_date = getattr(t, end_column)
        key = getattr(t, key_column)
        value = getattr(t, value_column)

        if end_date is pd.NaT:
            end_date = pd.Timestamp.today()

        first_month = start_date.normalize().to_period('M').to_timestamp('D', 'S')
        last_month = end_date.normalize().to_period('M').to_timestamp('D', 'S')

        index = pd.date_range(first_month, last_month, freq='MS')
        
        return pd.DataFrame(
            index=index,
            data=[[key]],
            columns=[value]
        )
    
    breakdown = pd.concat([build_df(t) for t in df.itertuples()]).resample('MS').agg(aggfunc)

    if output_columns:
        breakdown = breakdown[[s for s in output_columns if s in breakdown.columns]]
    
    return breakdown

def breakdown_by_month_sum_days(df, start_column, end_column, value_column, output_columns=None, aggfunc='sum'):
    """If `df` is a DataFrame of items that are valid/active between the
    timestamps stored in `start_column` and `end_column`, and where each has a
    categorical value in `value_column`, return a new DataFrame summing the
    overlapping days of items in each month broken down by each unique value in
    `value_column`. To restrict (and order) the value columns, pass a list of
    valid values as `output_columns`.
    """

    def build_df(t):
        start_date = getattr(t, start_column)
        end_date = getattr(t, end_column)
        value = getattr(t, value_column)

        if end_date is pd.NaT:
            end_date = pd.Timestamp.today()

        days_range = pd.date_range(start_date, end_date, freq='D')
        first_month = start_date.normalize().to_period('M').to_timestamp('D', 'S')
        last_month = end_date.normalize().to_period('M').to_timestamp('D', 'S')

        index = pd.date_range(first_month, last_month, freq='MS')
        
        return pd.DataFrame(
            index=index,
            data=[[len(pd.date_range(month_start, month_start + pd.tseries.offsets.MonthEnd(1), freq='D').intersection(days_range))] for month_start in index],
            columns=[value]
        )
    
    breakdown = pd.concat([build_df(t) for t in df.itertuples()]).resample('MS').agg(aggfunc)

    if output_columns:
        breakdown = breakdown[[s for s in output_columns if s in breakdown.columns]]
    
    return breakdown

def to_bin(value, edges):
    """Pass a list of numbers in `edges` and return which of them `value` falls
    between. If < the first item, return (0, <first>). If > last item, return
    (<last>, None).
    """

    previous = 0
    for v in edges:
        if previous <= value <= v:
            return (previous, v)
        previous = v
    return (previous, None)



import logging
import dateutil
import pandas as pd
from matplotlib import pyplot as plt

from ..calculator import Calculator
from ..utils import set_chart_style

logger = logging.getLogger(__name__)

class WasteCalculator(Calculator):
    """Calculate stories withdrawn, grouped by the time of withdrawal and
    stage prior to withdrawal.

    Chart is drawn in `waste_chart`, with title `waste_chart_title`, using
    tickets fetched with `waste_chart_query`, and grouped by month, limited to
    `waste_chart_window` months (if given).
    """

    def run(self):

        query = self.settings['waste_query']
        if not query:
            logger.debug("Not calculating waste chart data as no query specified")
            return None

        backlog_column = self.settings['backlog_column']
        done_column = self.settings['done_column']

        cycle_lookup = {}
        for idx, cycle_step in enumerate(self.settings['cycle']):
            for status in cycle_step['statuses']:
                cycle_lookup[status.lower()] = dict(
                    index=idx,
                    name=cycle_step['name'],
                    type=cycle_step['type'],
                )

        columns = ['key', 'last_status', 'resolution', 'withdrawn_date']
        series = {
            'key': {'data': [], 'dtype': 'str'},
            'last_status': {'data': [], 'dtype': 'str'},
            'resolution': {'data': [], 'dtype': 'str'},
            'withdrawn_date': {'data': [], 'dtype': 'datetime64[ns]'}
        }

        for issue in self.query_manager.find_issues(query):

            # Assume all waste items are resolved somehow
            if not issue.fields.resolution:
                continue
            
            last_status = None
            status_changes = list(self.query_manager.iter_changes(issue, ['status']))
            if len(status_changes) > 0:
                last_status = status_changes[-1].from_string
            
            if last_status is not None and last_status.lower() in cycle_lookup:
                last_status = cycle_lookup.get(last_status.lower())['name']
            else:
                logger.warn("Issue %s transitioned from unknown JIRA status %s", issue.key, last_status)

            # Skip if last_status was the backlog or done column (not really withdrawn)
            if last_status in (backlog_column, done_column):
                continue

            series['key']['data'].append(issue.key)
            series['last_status']['data'].append(last_status)
            series['resolution']['data'].append(issue.fields.resolution.name)
            series['withdrawn_date']['data'].append(dateutil.parser.parse(issue.fields.resolutiondate))

        data = {}
        for k, v in series.items():
            data[k] = pd.Series(v['data'], dtype=v['dtype'])

        return pd.DataFrame(data, columns=columns)

    def write(self):
        chart_data = self.get_result()
        if chart_data is None:
            return
        
        output_file = self.settings['waste_chart']
        if not output_file:
            logger.debug("No output file specified for waste chart")
            return
        
        if len(chart_data.index) == 0:
            logger.warning("Cannot draw waste chart with zero items")
            return

        frequency = self.settings['waste_frequency']
        window = self.settings['waste_window']
        
        cycle_names = [s['name'] for s in self.settings['cycle']]
        backlog_column = self.settings['backlog_column']
        done_column = self.settings['done_column']

        cycle_names.remove(backlog_column)
        cycle_names.remove(done_column)

        breakdown = chart_data \
                    .pivot_table(index='withdrawn_date', columns='last_status', values='key', aggfunc='count') \
                    .groupby(pd.Grouper(freq=frequency, closed='left', label='left')) \
                    .sum() \
                    .reindex(cycle_names, axis=1)

        if window:
            breakdown = breakdown[-window:]
        
        if len(breakdown.index) == 0 or len(breakdown.columns) == 0:
            logger.warning("Cannot draw waste chart with zero items")
            return

        fig, ax = plt.subplots()
        
        breakdown.plot.bar(ax=ax, stacked=True)
        
        if self.settings['waste_chart_title']:
            ax.set_title(self.settings['waste_chart_title'])

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        ax.set_xlabel("Month", labelpad=20)
        ax.set_ylabel("Number of items", labelpad=10)

        labels = [d.strftime("%b %y") for d in breakdown.index]
        ax.set_xticklabels(labels, rotation=90, size='small')

        set_chart_style()

        # Write file
        logger.info("Writing waste chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)
